#ifndef ASTGUARD
#define ASTGUARD

#include <string>
#include <list>

// **********************************************************************
// The ASTnode class defines the nodes of the abstract-syntax tree that
// represents a "Simple" program.
//
// Internal nodes of the tree contain pointers to children, organized
// either in a list (for nodes that may have a variable number of children)
// or as a fixed set of data members.  Leaf nodes contain line and column
// information; for string literals and identifiers, they also contain a
// string; for integer literals, they also contain an integer value.
//
// Here are all the different kinds of AST nodes and what kinds of children
// they have.  All of these kinds of AST nodes are subclasses of "ASTnode".
// Indentation indicates further subclassing:
//
//     Subclass            Kids
//     --------            ----
//     DeclListNode        list of DeclNode
//     FormalsListNode     list of FormalNode
//     FnBodyNode          DeclListNode, StmtListNode
//     StmtListNode        list of StmtNode
//     ExpListNode         list of ExpNode
//
//     DeclNode:
//       VarDeclNode       TypeNode, IdNode
//       FnDeclNode        IdNode, FormalsListNode, FnBodyNode
//
//     FormalNode          TypeNode, IdNode
//
//     TypeNode:
//       IntegerNode         -- none --
//       FloatNode         -- none --
//       BooleanNode         -- none --
//       StringNode          -- none --
//
//     StmtNode:
//       WriteStmtNode       ExpNode
//       AssignStmtNode      IdNode, ExpNode
//       IfStmtNode          ExpNode, StmtListNode
//       IfElseStmtNode      ExpNode, StmtListNode, StmtListNode
//       WhileStmtNode       ExpNode, StmtListNode
//       CallStmtNode        IdNode, ExpListNode
//       ReturnStmtNode      -- none --
//
//     ExpNode:
//       IntLitNode          -- none --
//       FloatLitNode        -- none --
//       StrLitNode          -- none --
//       TrueNode            -- none --
//       FalseNode           -- none --
//       IdNode              -- none --
//       UnaryMinusExpNode   ExpNode
//       NotExpNode          ExpNode
//       PlusExpNode         ExpNode, ExpNode
//       MinusExpNode        ExpNode, ExpNode
//       TimesExpNode        ExpNode, ExpNode
//       DivideExpNode       ExpNode, ExpNode
//       AndExpNode          ExpNode, ExpNode
//       OrExpNode           ExpNode, ExpNode
//       EqualsExpNode       ExpNode, ExpNode
//       NotEqualsExpNode    ExpNode, ExpNode
//       LessExpNode         ExpNode, ExpNode
//       GreaterExpNode      ExpNode, ExpNode
//       LessEqExpNode       ExpNode, ExpNode
//       GreaterEqExpNode    ExpNode, ExpNode
//
// Here are the different kinds of AST nodes again, organized according to
// whether they are leaves, internal nodes with lists of kids, or internal
// nodes with a fixed number of kids:
//
// (1) Leaf nodes:
//        IntegerNode, FloatNode, BooleanNode, StringNode, IntLitNode,
//	  FloatLitNode, StrLitNode, TrueNode, FalseNode, IdNode, ReturnStmtNode
//
// (2) Internal nodes with (possibly empty) lists of children:
//        FormalsListNode, DeclListNode, StmtListNode, ExpListNode
//
// (3) Internal nodes with fixed numbers of kids:
//        FnBodyNode,     DeclNode,      VarDeclNode,    FnDeclNode,
//        FormalNode,     WriteStmtNode,
//        AssignStmtNode, IfStmtNode,    IfElseStmtNode, WhileStmtNode,
//        CallStmtNode,   UnaryMinusExpNode,
//        NotExpNode,     PlusExpNode,   MinusExpNode,   TimesExpNode,
//        DivideExpNode,  AndExpNode,    OrExpNode,      EqualsExpNode,
//        NotEqualsExpNode,              LessExpNode,    GreaterExpNode,
//        LessEqExpNode,  GreaterEqExpNode 
//
// **********************************************************************

// **********************************************************************
// ASTnode class (base class for all other kinds of nodes)
// **********************************************************************
class ASTnode
{ 
 public:
  ASTnode() {};
  virtual void Unparse(int indent = 0) = 0; // every subtype must provide an Unparse operation
 private:
  ASTnode(const ASTnode & T);               // copy constructor hidden
  ASTnode & operator = (const ASTnode & T); // assignment hidden
};

// **********************************************************************
// subclasses of ASTnode having to do with statements and expressions
// **********************************************************************

class StmtNode: public ASTnode
{
};

class StmtListNode: public ASTnode
{
 public:
  StmtListNode(list<StmtNode *> *L);
  void Unparse(int indent=0);

 private:
  // list of kids
  list<StmtNode *> *myKids;
};

class ExpNode: public ASTnode
{
};

class ExpListNode: public ASTnode
{
 public:
  ExpListNode(list<ExpNode *> *L);
  void Unparse(int indent=0);

 private:
  // list of kids
  list<ExpNode *> *myKids;
};


// **********************************************************************
// Subclasses of ExpNode
// **********************************************************************

class IntLitNode: public ExpNode
{
 public:
  IntLitNode(int lineNum, int colNum, int intVal);
  void Unparse(int indent=0);

 private:
  int myLineNum;
  int myColNum;
  int myIntVal;
};

class FloatLitNode: public ExpNode
{
 public:
  FloatLitNode(int lineNum, int colNum, float floatVal);
  void Unparse(int indent=0);

 private:
  int myLineNum;
  int myColNum;
  float myFloatVal;
};

class StrLitNode: public ExpNode
{
 public:
  StrLitNode(int lineNum, int colNum, string *strVal);
  void Unparse(int indent=0);

 private:
  int myLineNum;
  int myColNum;
  string *myStrVal;
};

class TrueNode: public ExpNode
{
 public:
  TrueNode(int lineNum, int colNum);
  void Unparse(int indent=0);

 private:
  int myLineNum;
  int myColNum;
};

class FalseNode: public ExpNode
{
 public:
  FalseNode(int lineNum, int colNum);
  void Unparse(int indent=0);

 private:
  int myLineNum;
  int myColNum;
};

class IdNode: public ExpNode
{
 public:
  IdNode(int lineNum, int colNum, string *strVal);
  void Unparse(int indent=0);

 private:
  int myLineNum;
  int myColNum;
  string *myStrVal;
};

// Unary operators -----------------------------------
class UnaryMinusExpNode: public ExpNode
{
 public:
  UnaryMinusExpNode(ExpNode *exp);
  void Unparse(int indent=0);
 private:
  // 1 kid
  ExpNode *myExp;
};

class NotExpNode: public ExpNode
{
 public:
  NotExpNode(ExpNode *exp);
  void Unparse(int indent=0);
 private:
  // 1 kid
  ExpNode *myExp;
};

// Binary operators -----------------------------------
class PlusExpNode: public ExpNode
{
 public:
  PlusExpNode(ExpNode *exp1, ExpNode *exp2);
  void Unparse(int indent=0);
 private:
  // 2 kids
  ExpNode *myExp1;
  ExpNode *myExp2;
};

class MinusExpNode: public ExpNode
{
 public:
  MinusExpNode(ExpNode *exp1, ExpNode *exp2);
  void Unparse(int indent=0);
 private:
  // 2 kids
  ExpNode *myExp1;
  ExpNode *myExp2;
};

class TimesExpNode: public ExpNode
{
 public:
  TimesExpNode(ExpNode *exp1, ExpNode *exp2);
  void Unparse(int indent=0);
 private:
  // 2 kids
  ExpNode *myExp1;
  ExpNode *myExp2;
};

class DivideExpNode: public ExpNode
{
 public:
  DivideExpNode(ExpNode *exp1, ExpNode *exp2);
  void Unparse(int indent=0);
 private:
  // 2 kids
  ExpNode *myExp1;
  ExpNode *myExp2;
};

class AndExpNode: public ExpNode
{
 public:
  AndExpNode(ExpNode *exp1, ExpNode *exp2);
  void Unparse(int indent=0);
 private:
  // 2 kids
  ExpNode *myExp1;
  ExpNode *myExp2;
};

class OrExpNode: public ExpNode
{
 public:
  OrExpNode(ExpNode *exp1, ExpNode *exp2);
  void Unparse(int indent=0);
 private:
  // 2 kids
  ExpNode *myExp1;
  ExpNode *myExp2;
};

class EqualsExpNode: public ExpNode
{
 public:
  EqualsExpNode(ExpNode *exp1, ExpNode *exp2);
  void Unparse(int indent=0);
 private:
  // 2 kids
  ExpNode *myExp1;
  ExpNode *myExp2;
};

class NotEqualsExpNode: public ExpNode
{
 public:
  NotEqualsExpNode(ExpNode *exp1, ExpNode *exp2);
  void Unparse(int indent=0);
 private:
  // 2 kids
  ExpNode *myExp1;
  ExpNode *myExp2;
};

class LessExpNode: public ExpNode
{
 public:
  LessExpNode(ExpNode *exp1, ExpNode *exp2);
  void Unparse(int indent=0);
 private:
  // 2 kids
  ExpNode *myExp1;
  ExpNode *myExp2;
};

class GreaterExpNode: public ExpNode
{
 public:
  GreaterExpNode(ExpNode *exp1, ExpNode *exp2);
  void Unparse(int indent=0);
 private:
  // 2 kids
  ExpNode *myExp1;
  ExpNode *myExp2;
};

class LessEqExpNode: public ExpNode
{
 public:
  LessEqExpNode(ExpNode *exp1, ExpNode *exp2);
  void Unparse(int indent=0);
 private:
  // 2 kids
  ExpNode *myExp1;
  ExpNode *myExp2;
};

class GreaterEqExpNode: public ExpNode
{
 public:
  GreaterEqExpNode(ExpNode *exp1, ExpNode *exp2);
  void Unparse(int indent=0);
 private:
  // 2 kids
  ExpNode *myExp1;
  ExpNode *myExp2;
};

// **********************************************************************
// Subclasses of StmtNode
// **********************************************************************
class WriteStmtNode: public StmtNode
{
 public:
  WriteStmtNode(ExpNode *exp);
  void Unparse(int indent=0);

 private:
  // 1 kid
  ExpNode * myExp;
};

class AssignStmtNode: public StmtNode
{
 public:
  AssignStmtNode(IdNode *id, ExpNode *exp);
  void Unparse(int indent=0);

 private:
  // 2 kids
  IdNode * myId;
  ExpNode * myExp;
};

class IfStmtNode: public StmtNode
{
 public:
  IfStmtNode(ExpNode *exp, StmtListNode *stmtList);
  void Unparse(int indent=0);

 private:
  // 2 kids
  ExpNode * myExp;
  StmtListNode * myStmtList;
};

class IfElseStmtNode: public StmtNode
{
 public:
  IfElseStmtNode(ExpNode *exp, StmtListNode *thenStmtList,
		 StmtListNode *elseStmtList);
  void Unparse(int indent=0);

 private:
  // 3 kids
  ExpNode * myExp;
  StmtListNode * myThenStmtList;
  StmtListNode * myElseStmtList;
};

class WhileStmtNode: public StmtNode
{
 public:
  WhileStmtNode(ExpNode *exp, StmtListNode *stmtList);
  void Unparse(int indent=0);

 private:
  // 2 kids
  ExpNode * myExp;
  StmtListNode * myStmtList;
};

class CallStmtNode: public StmtNode
{
 public:
  CallStmtNode(IdNode *id, ExpListNode *expList);
  void Unparse(int indent=0);

 private:
  // 2 kids
  IdNode * myId;
  ExpListNode * myExpList;
};

class ReturnStmtNode: public StmtNode
{
 public:
  ReturnStmtNode(int lineNum, int colNum);
  void Unparse(int indent=0);            

 private:
  int myLineNum;
  int myColNum;
};

// **********************************************************************
// TypeNode and its Subclasses
// **********************************************************************
class TypeNode: public ASTnode
{
};

class IntegerNode: public TypeNode
{
 public:
  IntegerNode(int lineNum, int colNum);
  void Unparse(int indent=0);

 private:
  int myLineNum;
  int myColNum;
};

class FloatNode: public TypeNode
{
 public:
  FloatNode(int lineNum, int colNum);
  void Unparse(int indent=0);

 private:
  int myLineNum;
  int myColNum;
};

class BooleanNode: public TypeNode
{
 public:
  BooleanNode(int lineNum, int colNum);
  void Unparse(int indent=0);

 private:
  int myLineNum;
  int myColNum;
};

class StringNode: public TypeNode
{
 public:
  StringNode(int lineNum, int colNum);
  void Unparse(int indent=0);

 private:
  int myLineNum;
  int myColNum;
};

// **********************************************************************
// FormalNode and its Subclasses
// **********************************************************************
class FormalNode: public ASTnode
{
 public:
  FormalNode(TypeNode *type, IdNode * id);
  void Unparse(int indent=0);

 private:
  // 2 kids
  TypeNode * myType;
  IdNode * myId;
};

// **********************************************************************
// more subclasses of ASTnode:
//   DeclNode, DeclListNode, FnBodyNode, and FormalsListNode
// **********************************************************************
class DeclNode: public ASTnode
{
};

class DeclListNode: public ASTnode
{
 public:
  DeclListNode(list<DeclNode *> *L);
  void Unparse(int indent=0);

 private:
  // list of kids
  list<DeclNode *> *myKids;
};

class FnBodyNode: public ASTnode
{
 public:
  FnBodyNode(DeclListNode *declList, StmtListNode * stmtList);
  void Unparse(int indent=0);

 private:
  // 2 kids
  DeclListNode * myDeclList;
  StmtListNode * myStmtList;
};

class FormalsListNode: public ASTnode
{
 public:
  FormalsListNode(list<FormalNode *> *L);
  void Unparse(int indent=0);

 private:
  // list of kids
  list<FormalNode *> *myKids;
};


// **********************************************************************
// Subclasses of Declnode
// **********************************************************************
class VarDeclNode: public DeclNode
{
 public:
  VarDeclNode(TypeNode *type, IdNode *id);
  void Unparse(int indent=0);

 private:
  // 2 kids
  TypeNode * myType;
  IdNode * myId;
};

class FnDeclNode: public DeclNode
{
 public:
  FnDeclNode(IdNode *id, FormalsListNode *formalList, FnBodyNode *body);
  void Unparse(int indent=0);

 private:
  // 3 kids
  IdNode * myId;
  FormalsListNode * myFormalsList;
  FnBodyNode * myBody;
};

// **********************************************************************
// GLOBAL variable: the root of the AST built by a (successful) parse)
// **********************************************************************
extern DeclListNode *AST;

#endif
