/**********************************************************************
 Parser for Simple programs
 **********************************************************************/

// The types of the tokens and nonterminals
%union {
  TokenType         TOKENval;   // type of all tokens
  DeclNode *        DeclNodeval;
  TypeNode *        TypeNodeval;
  IdNode *          IdNodeval;
  FormalsListNode *  FormalVal;
  FnBodyNode *	    FnBodyval;
  ExpNode *	    ExpNodeval;
  StmtNode * 	    StmtNodeval;
  list<DeclNode *> *DeclListval;
  list<FormalNode *> *FormalListval;
  list<StmtNode *>   *StmtListval;
  list<ExpNode *>   *ExpListval;
};

%token <TOKENval> INTEGER
%token <TOKENval> FLOAT
%token <TOKENval> BOOLEAN
%token <TOKENval> STRING
%token <TOKENval> VOID
%token <TOKENval> TRUE
%token <TOKENval> FALSE
%token <TOKENval> IF
%token <TOKENval> ELSE
%token <TOKENval> WHILE
%token <TOKENval> DO
%token <TOKENval> WRITE
%token <TOKENval> RETURN
%token <TOKENval> ID
%token <TOKENval> INTLITERAL
%token <TOKENval> FLOATLITERAL
%token <TOKENval> STRINGLITERAL
%token <TOKENval> LCURLY
%token <TOKENval> RCURLY
%token <TOKENval> LPAREN
%token <TOKENval> RPAREN
%token <TOKENval> COMMA
%token <TOKENval> ASSIGN
%token <TOKENval> SEMICOLON
%token <TOKENval> PLUS
%token <TOKENval> MINUS
%token <TOKENval> TIMES
%token <TOKENval> DIVIDE
%token <TOKENval> NOT
%token <TOKENval> AND
%token <TOKENval> OR
%token <TOKENval> EQUALS
%token <TOKENval> NOTEQUALS
%token <TOKENval> LESS
%token <TOKENval> GREATER
%token <TOKENval> LESSEQ
%token <TOKENval> GREATEREQ

%left AND OR
%nonassoc EQUALS NOTEQUALS LESS GREATER LESSEQ GREATEREQ
%left PLUS MINUS
%left TIMES DIVIDE
%left NOT

%type <DeclListval> program  declList varDeclList 
%type <DeclNodeval> decl varDecl fnDecl
%type <TypeNodeval> type
%type <FormalListval> formalList
%type <StmtListval>  stmtList
%type <StmtNodeval> stmt
%type <FormalVal> formals
%type <IdNodeval>   id 
%type <FnBodyval>   fnBody
%type <ExpNodeval>  exp
%type <ExpListval>  actualList
 
%start program

%%
program		: declList
                  { AST = new DeclListNode($1);
		  }
		;

declList 	: declList decl
                  { $$ = $1;
		    $$->push_back($2);
		  }
		| /* epsilon */
                  { $$ = new list<DeclNode *>; }
		;

decl		: varDecl
                  { $$ = $1; }
		| fnDecl
                  { $$ = $1; }
		;
varDecl		: type id SEMICOLON
                  { $$ = new VarDeclNode($1, $2); }
		;


fnDecl		: VOID id formals fnBody 
		{ $$ = new FnDeclNode($2, $3, $4);}
		;
type		: INTEGER
                  { $$ = new IntegerNode($1.lineNum, $1.colNum); }
	        | FLOAT
                  { $$ = new FloatNode($1.lineNum, $1.colNum); }
	        | BOOLEAN
                  { $$ = new BooleanNode($1.lineNum, $1.colNum); }
                | STRING
                  { $$ = new StringNode($1.lineNum, $1.colNum); }
		;

formals		: LPAREN RPAREN
		{ $$ = 0; }
		| LPAREN formalList RPAREN
		{ $$ = new FormalsListNode($2); } 
		;

formalList	: type id
                {
			$$ = new list<FormalNode *>;
			$$->push_back(new FormalNode($1, $2));
		}
		| formalList COMMA type id
		{
			$$ = $1;
			$$->push_back(new FormalNode($3, $4) );
		}
		;

fnBody		: LCURLY varDeclList  stmtList RCURLY
		{ $$ = new FnBodyNode(new DeclListNode($2), new StmtListNode($3)); };

varDeclList 	: varDeclList varDecl
                  { $$ = $1;
		    $$->push_back($2);
		  }
		| /* epsilon */
                  { $$ = new list<DeclNode *>; }
		;

stmtList	: stmtList stmt
		{
			$$ = $1;
			$$->push_back($2);
		}
		| /* epsilon */
	 	{ $$ = new list<StmtNode *>; }
		;

stmt		: WRITE LPAREN exp RPAREN SEMICOLON
		{
			$$ = new WriteStmtNode($3);
		}
		| id ASSIGN exp SEMICOLON
		{
			$$ = new AssignStmtNode($1, $3);
		}
		| IF LPAREN exp RPAREN LCURLY stmtList RCURLY
		{
			$$ = new IfStmtNode($3, new StmtListNode($6));
		}
		| IF LPAREN exp RPAREN LCURLY stmtList RCURLY ELSE LCURLY stmtList RCURLY
		{
			$$ = new IfElseStmtNode($3, new StmtListNode($6), new StmtListNode($10));
		
		}
		| WHILE LPAREN exp RPAREN LCURLY stmtList RCURLY 
		{
			$$ = new WhileStmtNode($3, new StmtListNode($6));
		}
		| RETURN SEMICOLON
		{
			$$ = new ReturnStmtNode($1.lineNum, $1.colNum);
		}
		| id LPAREN RPAREN SEMICOLON
		{
			$$ = new CallStmtNode($1, (ExpListNode *)0);
		}
		| id LPAREN actualList RPAREN SEMICOLON
		{
			$$ = new CallStmtNode($1, new ExpListNode($3));
		}
		;

exp		: id
		{	$$ = $1; }
		| INTLITERAL
		{ $$ = new IntLitNode($1.lineNum, $1.colNum, $1.intVal); }
		| FLOATLITERAL
		{ $$ = new FloatLitNode($1.lineNum, $1.colNum, $1.floatVal); }
		| STRINGLITERAL 
		{ $$ = new StrLitNode($1.lineNum, $1.colNum, $1.strVal); }
		| TRUE 
		{ $$ = new TrueNode($1.lineNum, $1.colNum); }
		| FALSE 
		{ $$ = new FalseNode($1.lineNum, $1.colNum); }
		| MINUS exp %prec NOT
		{ $$ = new UnaryMinusExpNode($2); }
		| exp PLUS exp
		{ $$ = new PlusExpNode($1, $3); }
		| exp MINUS exp
		{ $$ = new MinusExpNode($1, $3); }
		| exp TIMES exp
		{ $$ = new TimesExpNode($1, $3); }
		| exp DIVIDE exp
		{ $$ = new DivideExpNode($1, $3); }
		| NOT exp
		{ $$ = new NotExpNode($2); }
		| exp AND exp
		{ $$ = new AndExpNode($1, $3); }
		| exp OR exp
		{ $$ = new OrExpNode($1, $3); }
		| exp EQUALS exp
		{ $$ = new EqualsExpNode($1, $3); }
		| exp NOTEQUALS exp
		{ $$ = new NotEqualsExpNode($1, $3); }
		| exp LESS exp
		{ $$ = new LessExpNode($1, $3); }
		| exp GREATER exp
		{ $$ = new GreaterExpNode($1, $3); }
		| exp LESSEQ exp
		{ $$ = new LessEqExpNode($1, $3); }
		| exp GREATEREQ exp
		{ $$ = new GreaterEqExpNode($1, $3); }
		| LPAREN exp RPAREN %prec NOT
		{ $$ = $2; }
		;
		
actualList	: exp
		{  
			$$ = new list<ExpNode *>;
			$$->push_back($1);
		}
		  
		| actualList COMMA exp
		{
			$$ = $1;
			$$->push_back($3);
		}
		;

id              : ID
                  { $$ = new IdNode($1.lineNum, $1.colNum, $1.strVal); }

%%

